// Copyright (C) 1994-98 by Addison-Wesley Publishing Co., Inc.
//
// This code is provided for its instructional value.
// It has been tested with care, but it is not guaranteed
// for any particular purpose.  The publisher and authors
// do not offer any warranties or representations, nor do
// they accept any liabilities with respect to this code.

#include "cpp-faqs.hpp"

#include <cassert>
#include <stdexcept>
using namespace std;

class Fred {
public:
  Fred() throw();                       // A default constructor
  Fred(int i, int j) throw();           // A normal constructor
  Fred(const Fred& f) throw();
  Fred& operator= (const Fred& f) throw();
 ~Fred() throw();
  void sampleInspectorMethod() const throw();  // Does not change this object
  void sampleMutatorMethod() throw(bad_alloc); // Changes this object
  // ...

private:

  class Data {
  public:
    Data() throw();
    Data(int i, int j) throw();
    Data(const Data& d) throw();

    // Since only Fred can access Fred::Data,
    // the Fred::Data members can be public:.
    // If that feels uncomfortable, the members can be made private:
    // and class Fred can be made a friend class. See FAQ 19.01
    // ...

    unsigned count_;
    // count_ is the number of Fred objects that point at this
    // count_ must be initialized to 1 by all constructors
    // (starts at 1 because of the Fred object that created this)
  };

  Data* data_;
};

Fred::Data::Data() throw()              : count_(1) /*...*/ { }
Fred::Data::Data(int i, int j) throw()  : count_(1) /*...*/ { }
Fred::Data::Data(const Data& d) throw() : count_(1) /*...*/ { }

Fred::Fred() throw()              : data_(new Data()) { }
Fred::Fred(int i, int j) throw()  : data_(new Data(i, j)) { }
Fred::Fred(const Fred& f) throw() : data_(f.data_) { ++data_->count_; }

Fred& Fred::operator= (const Fred& f) throw()
{
  // DO NOT CHANGE THE ORDER OF THESE STATEMENTS!
  // (This order properly handles self-assignment; see FAQ 24.02)
  ++f.data_->count_;
  if (--data_->count_ == 0) delete data_;
  data_ = f.data_;
  return *this;
}

Fred::~Fred() throw()
{ if (--data_->count_ == 0) delete data_; }

void Fred::sampleInspectorMethod() const throw()
{
  // This member function promises ("const") not to change anything in *data_
  // Other than that, any data access would simply use "data_->..."
}

void Fred::sampleMutatorMethod() throw(bad_alloc)
{
  // This member function might need to change things in *data_
  // Thus it first checks if this is the only pointer to *data_
  if (data_->count_ > 1) {
    Data* d = new Data(*data_);    // Invoke Fred::Data's copy ctor
    -- data_->count_;
    data_ = d;
  }
  assert(data_->count_ == 1);

  // Now the member function proceeds to access "data_->..." as normal
}
